
#include <boost/property_tree/ptree.hpp>

#ifdef MUQ_MPI
# include <mpi.h>
# include <boost/mpi.hpp>
namespace mpi = boost::mpi;
#endif // if MUQ_MPI


#include "MUQ/Inference/TransportMaps/TransportMap.h"
#include "MUQ/Inference/TransportMaps/MapFactory.h"
#include "MUQ/Inference/TransportMaps/LinearBasis.h"
#include "MUQ/Inference/TransportMaps/StaticPolynomials.h"
#include "MUQ/Inference/TransportMaps/PolynomialBasis.h"
#include "MUQ/Inference/TransportMaps/BasisFactory.h"

#include "MUQ/Inference/MCMC/SingleChainMCMC.h"
#include "MUQ/Inference/MCMC/TransportMapKernel.h"
#include "MUQ/Inference/ProblemClasses/InferenceProblem.h"
#include "MUQ/Inference/MAP/MAPbase.h"

#include "MUQ/Inference/ProblemClasses/AbstractMetric.h"

#include "MUQ/Utilities/JobManager.h"
#include "MUQ/Utilities/HDF5Wrapper.h"

#include "MUQ/Modelling/DensityProduct.h"
#include "MUQ/Modelling/EmpiricalRandVar.h"
#include "MUQ/Modelling/ModGraph.h"
#include "MUQ/Modelling/GaussianPair.h"
#include "MUQ/Modelling/VectorPassthroughModel.h"

#include <boost/timer/timer.hpp>

#include <fstream>
#include <iostream>

/// namespaces
using namespace muq::Modelling;
using namespace muq::Inference;
using namespace muq::Utilities;
using namespace std;

class fcache{
  
public:
  
  void Update(const Eigen::VectorXd &w){
    if(w!=wSaved){
      wSaved = w;
      f = (*predictors)*wSaved;
      fExp = f.array().exp().matrix();
    }
  };
  
  Eigen::VectorXd wSaved;
  Eigen::VectorXd f;
  Eigen::Array<double,Eigen::Dynamic,1> fExp;
  std::shared_ptr<Eigen::MatrixXd> predictors; 
  
};
class GermanCreditLikelihood : public Density{
  
public:
  
  
  GermanCreditLikelihood(const std::string &dataFile) : Density(Eigen::VectorXi::Constant(1,25),false,false,false,false), numEvals(0), numGradEvals(0){
    
    std::cout << "Reading input data... " << std::flush;
    boost::timer::cpu_timer timer;
    
    // open the data file
    const int numAttributes = 24;
    std::string temp;
    
    // get the number of lines in the file
    ifstream fin(dataFile.c_str());
    fin.unsetf(std::ios_base::skipws);
    numData = std::count(istream_iterator<char>(fin), istream_iterator<char>(), '\n');
    fin.clear();
    fin.seekg(0,ios::beg);
    
    // resize the data
    predictors = make_shared<Eigen::MatrixXd>(numData,numAttributes+1);
    predictors->col(0) = Eigen::VectorXd::Ones(numData);
    types.resize(numData);
    
    // read in the predictors and cover types
    for(int i=0; i<numData;  ++i){
      for(int k=0; k<numAttributes; ++k){
        fin >> std::skipws >> temp;
        (*predictors)(i,k+1) = stof(temp);
      }
      
      // now get the cover type
      getline(fin,temp,'\n');
      types(i) = stoi(temp)-1.0;
      assert(types(i)<2);
      
    }
    
    fin.close();
    
    // scale all the predictors to have mean zero and variance 1
    Eigen::RowVectorXd mu = predictors->colwise().sum()/numData;
    Eigen::MatrixXd diff = predictors->rowwise()-mu;
    Eigen::RowVectorXd std = (diff.array().pow(2.0).matrix().colwise().sum()/(numData-1.0)).array().sqrt();
    for(int i=1; i<numAttributes+1; ++i)
      predictors->col(i) = diff.col(i)/std(i);
    
    boost::timer::cpu_times elapsed = timer.elapsed();
    double elapsedMilli = double(elapsed.user + elapsed.system)/1e6;
    std::cout << "done in " << elapsedMilli << " milliseconds.\n";
    
    predictCache.predictors = predictors;
    predictCache.wSaved = Eigen::VectorXd::Random(predictors->cols());
    predictCache.Update(Eigen::VectorXd::Zero(predictors->cols()));
    
  };
  
  
  
  int GetNumEvals() const{return numEvals;};
  int GetNumGradEvals() const{return numGradEvals;};
  
  
  virtual double LogDensityImpl(std::vector<Eigen::VectorXd> const& input) override{
    numEvals++;
    predictCache.Update(input.at(0));
    double logLikely = types.dot(predictCache.f) - (1.0+predictCache.fExp).log().sum();
    return logLikely;
  }
  
  virtual Eigen::VectorXd GradientImpl(std::vector<Eigen::VectorXd> const& input, Eigen::VectorXd const& sens, int inputDimWrt) override{
    numGradEvals++;
    predictCache.Update(input.at(0));
    return sens(0)*predictors->transpose()*(types.array() - predictCache.fExp/(1.0+predictCache.fExp)).matrix();
  }
  
  std::shared_ptr<Eigen::MatrixXd> predictors;
  fcache predictCache;
  
  Eigen::VectorXd types;
  int numData;
  int numEvals, numGradEvals;
  
};


class LogisticMetric : public AbstractMetric{
  
public:
  LogisticMetric(std::shared_ptr<GermanCreditLikelihood> likelyIn, double priorVarIn) : likelihood(likelyIn), priorVar(priorVarIn){};
  
  virtual Eigen::MatrixXd ComputeMetric(Eigen::VectorXd const& point){
    likelihood->predictCache.Update(point);
    Eigen::VectorXd p = (1.0+(-1.0*likelihood->predictCache.f).array().exp()).inverse();
    Eigen::VectorXd v = p.array()*(1.0-p.array());
    Eigen::MatrixXd G = likelihood->predictors->transpose()*v.asDiagonal()*(*likelihood->predictors) + (1.0/priorVar)*Eigen::MatrixXd::Identity(point.size(),point.size());
    return G;
  };
  
private:
  std::shared_ptr<GermanCreditLikelihood> likelihood;
  double priorVar;
};

void ReadParams(const string &filename, boost::property_tree::ptree &params){
  
  // open a filestream
  ifstream fin(filename.c_str());
  boost::property_tree::read_xml(fin,params);
  fin.close();
}


string GetHdf5Root(boost::property_tree::ptree const& params)
{
  std::stringstream ss;
  bool hasCopyNumber = params.get("CopyNumber",-1)!=-1;
  
  if(hasCopyNumber)
    ss << std::setfill ('0') << std::setw (3) << params.get<int>("CopyNumber");
  
  string copyNum = ss.str();
  
  string rootName;
  string problemName = params.get("ProblemName","");
  string kernelName = params.get<std::string>("MCMC.Kernel");
  
  if(problemName.compare("")!=0)
    rootName = "/" + problemName;
  
  if((problemName.compare("BOD")==0)&&(kernelName.compare("TransportMap")==0))  
    rootName += "/PolyOrder" + params.get<std::string>("MCMC.TransportMap.PolyOrder");
  
  rootName += "/" + kernelName;
  
  if(kernelName.compare("MHKernel")==0)
    rootName += "/" + params.get<std::string>("MCMC.Proposal");
  
  if(kernelName.compare("DR")==0)
    rootName += "/" + params.get<std::string>("MCMC.DR.ProposalSpecification");
  
  if(kernelName.compare("TransportMap")==0)
  {
    string subRoot = GetHdf5Root(params.get_child("MCMC.TransportMap.SubMethod"));
    rootName += subRoot;
  }
  
  if(hasCopyNumber)
    rootName += "/Trial" + copyNum;
  
  return rootName;
  
}

void WriteResults(const string &datasetRoot, boost::property_tree::ptree &params, shared_ptr<EmpiricalRandVar> samples, int numEvals, int numGradEvals, double runTime){
  
  std::cout << "Writing results into " << datasetRoot << std::endl;
  samples->WriteHDF5(datasetRoot+ "/chain");
  
  HDF5Wrapper::WriteMatrix<double,Eigen::Dynamic,1>(datasetRoot + "/evals", Eigen::VectorXd::Constant(1,numEvals));
  HDF5Wrapper::WriteMatrix<double,Eigen::Dynamic,1>(datasetRoot + "/gradEvals", Eigen::VectorXd::Constant(1,numGradEvals));
  HDF5Wrapper::WriteMatrix<double,Eigen::Dynamic,1>(datasetRoot + "/runtime", Eigen::VectorXd::Constant(1,runTime));
}


void RunGermanCredit(boost::property_tree::ptree const& constParams){
  
  
  assert(constParams.get("CopyNumber",-1)!=-1);
  boost::property_tree::ptree params;
  params.put_child("MCMC",constParams.get_child("MCMC"));
  params.put_child("Opt",constParams.get_child("Opt"));
  
  params.put("ProblemName","GermanCredit");
  params.put("Verbose",3);
  params.put("CopyNumber",constParams.get<int>("CopyNumber"));
  
  // get the hdf root group
  string hdf5Root = GetHdf5Root(params);
  
  std::cout << "Running: " << hdf5Root << std::endl;
  const int probDim = 25;
  
  auto likelihoodDens = make_shared<GermanCreditLikelihood>("inputParams/german.data-numeric");
  
  const double priorVar = 100.0;
  auto priorDens = make_shared<GaussianDensity>(Eigen::VectorXd::Zero(probDim),priorVar);
  
  
  shared_ptr<ModGraph> sys = make_shared<ModGraph>();
  sys->AddNode(make_shared<VectorPassthroughModel>(probDim),"inferenceTarget");
  
  // add the prior to the graph
  sys->AddNode(priorDens,"prior");
  sys->AddEdge("inferenceTarget","prior",0);
  
  // add the likelihood to the graph
  sys->AddNode(likelihoodDens,"likelihood");
  sys->AddEdge("inferenceTarget","likelihood",0);
  
  // create the posterior
  sys->AddNode(make_shared<DensityProduct>(2), "posterior");	
  sys->AddEdge("prior","posterior",0);
  sys->AddEdge("likelihood","posterior",1);
  
  
  // create the inference problem
  std::shared_ptr<InferenceProblem> problem;
  if(params.get<std::string>("MCMC.Proposal").compare("MMALA")){
    problem = make_shared<InferenceProblem>(sys);
  }else{
    
    // create the metric
    auto metric = make_shared<LogisticMetric>(likelihoodDens,priorVar);
    
    // now create the problem
    problem = make_shared<InferenceProblem>(sys,metric);
  }
  
  
  // read in the solver parameters
  //boost::property_tree::ptree params;
  //ReadParams("inputParams/GermanLogisticParams.xml",params);
  
  // set an initial point
  Eigen::VectorXd beta0 = Eigen::VectorXd::Zero(probDim);
  
  // find the map point 
  auto mapSolver = MAPbase::Create(problem,params);
  Eigen::VectorXd betaMAP = mapSolver->Solve(beta0);
  
  
  // create the mcmc solver
  std::shared_ptr<MCMCBase> solver;
  if(params.get<std::string>("MCMC.Kernel").compare("TransportMap"))
  {
    solver = MCMCBase::ConstructMCMC(problem, params);	
  }
  else
  {
    auto bases = BasisFactory::TotalOrderHermite(probDim, params.get("MCMC.TransportMap.GlobalPolyOrder",1));
    bases += BasisFactory::NoCrossHermite(probDim,params.get("MCMC.TransportMap.LocalPolyOrder",1));	
    auto map = MapFactory::CreateIdentity(bases);
    
    std::cout << "Using TransportMapMCMC\n";
    auto TestKernel = make_shared<TransportMapKernel>(problem,params,map);
    solver = make_shared<SingleChainMCMC>(TestKernel,params);
  }
  
  boost::timer::cpu_timer timer;
  
  // sample the posterior
  auto mcmcSamps = solver->Sample(betaMAP);
  
  // write the results
  boost::timer::cpu_times elapsed = timer.elapsed();
  double runTime = double(elapsed.user + elapsed.system)/1e6; // cpu time in milliseconds
  
  WriteResults(hdf5Root, params, mcmcSamps, likelihoodDens->GetNumEvals(),likelihoodDens->GetNumGradEvals(),runTime);
  solver->WriteAttributes(hdf5Root); // needs to be called after groups have been created
  
  HDF5Wrapper::FlushFile();
  
}

REGISTER_JOB_FUNCTION(RunGermanCredit, GetHdf5Root)


int main(int argc, char **argv){
  
  #ifdef MUQ_MPI
  std::cout << "Intializing MUQ_MPI...\n";
  boost::mpi::environment env(argc, argv);
  #endif // ifdef MUQ_MPI
  
  std::vector<string> MCMCKernels, TransportMapKernels, TransportMapProposals, MCMCProposals;
  
  // // job 1 -- NUTS
  MCMCKernels.push_back("StanNUTS");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 2 -- DR
  MCMCKernels.push_back("DR");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  //  job 3 -- AMALA
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("AMALA");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 4 - smMALA
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("MMALA");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 5 -- TMDRG
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DR");
  TransportMapProposals.push_back("GlobalProposal,LocalProposal1");
  
  // job 6 -- TMDRL
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DR");
  TransportMapProposals.push_back("LocalProposal2,LocalProposal3");
  
  // job 7 -- TMMIX
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("MixtureProposal");
  
  // job 8 -- TMRWM
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("MHProposal");
  
  // job 8 -- TMLA
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("PreMALA");
  
  
  const int numRepeats = 30;
  
  boost::property_tree::ptree jobConfig;
  
  /////////////////////////////////////////////////////
  // set up all the BOD jobs
  boost::property_tree::ptree localJob;
  localJob.put("NumberOfCopies", numRepeats);
  localJob.put("Function", "RunGermanCredit");
  
  
  // open up the hdf5 file
  HDF5Wrapper::OpenFile("results/GermanCreditResults.h5");
  
  for(int i=0; i<MCMCKernels.size(); ++i){
    
    boost::property_tree::ptree mcmcParams;
    ReadParams("inputParams/GermanLogisticParams.xml",mcmcParams);
    
    mcmcParams.put("MCMC.Kernel",MCMCKernels.at(i));
    mcmcParams.put("MCMC.Proposal",MCMCProposals.at(i));
    
    // if this is a transport map, add a job for each polynomial order
    if(TransportMapKernels.at(i).compare("DUMMY")){
      
      mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.Kernel",TransportMapKernels.at(i));
      
      if(TransportMapKernels.at(i).compare("DR")==0){
        mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.DR.ProposalSpecification",TransportMapProposals.at(i));
      }else{
        mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.Proposal",TransportMapProposals.at(i));        
      }
    }
    
    localJob.put_child("Data",mcmcParams);
    jobConfig.add_child("Job",localJob);        
    
  }
  
  JobManager::RunJobs(jobConfig); 
  
  std::cout << "About to enter CloseFile()" << std::endl;
  
  // close the results file
  HDF5Wrapper::CloseFile();      
  
  
  return 0;
}
