
//self include
#include "PopulationModels.h"

// boost includes
#include <boost/property_tree/ptree.hpp>

// MUQ includes
#include "MUQ/Modelling/OdeModPiece.h"

#include <Eigen/Dense>

using namespace ecology;
using namespace muq::Modelling;
using namespace std;

PredPreyRHS::PredPreyRHS() : ModPiece(Eigen::Vector3i{1, 2, 6}, 2,
                          	 		  false,    // does this class provide an efficient implementation of the GradientImpl function?
                           		   	  true,    // does this class provide an efficient implementation of the JacobianImpl function?
                           		   	  false,    // does this class provide an efficient implementation of the JacobianActionImpl function?
                           		   	  false,    // does this class provide an efficient implementation of the HessianImpl function?
                           		   	  false) {};// does this class return a random result?
									  
									  
									  
Eigen::VectorXd PredPreyRHS::EvaluateImpl(std::vector<Eigen::VectorXd> const& inputs)
{
  // grad the current time in the simulation
  double time = inputs.at(0)(0);

  // for clarity, grab the prey and predator populations from the input vector
  double preyPop = inputs.at(1)(0);
  double predPop = inputs.at(1)(1);


  double preyGrowth, preyCapacity, predationRate, predationHunger, predatorGrowth, predatorLoss;

  preyGrowth      = inputs.at(2)(0);
  preyCapacity    = inputs.at(2)(1);
  predationRate   = inputs.at(2)(2);
  predationHunger = inputs.at(2)(3);
  predatorGrowth  = inputs.at(2)(4);
  predatorLoss    = inputs.at(2)(5);

  // create a vector to hold the ModPiece output, this will hold the time derivatives of population
  Eigen::VectorXd output(2);

  // the first output of this function is the derivative of the predator population size with resepct to time
  double preyEaten = predationRate * preyPop * predPop / (predationHunger + preyPop);
  output(0) = preyGrowth * preyPop * (1 - preyPop / preyCapacity) - preyEaten;
  output(1) = predatorGrowth * preyEaten - predatorLoss * predPop;

  // return the output
  return output;
}


Eigen::MatrixXd PredPreyRHS::JacobianImpl(std::vector<Eigen::VectorXd> const& inputs, int inputDimWrt)
{
  assert(inputDimWrt!=0);
  
    // for clarity, grab the prey and predator populations from the input vector
    double preyPop = inputs.at(1)(0);
    double predPop = inputs.at(1)(1);


    double preyGrowth, preyCapacity, predationRate, predationHunger, predatorGrowth, predatorLoss;

    preyGrowth      = inputs.at(2)(0);
    preyCapacity    = inputs.at(2)(1);
    predationRate   = inputs.at(2)(2);
    predationHunger = inputs.at(2)(3);
    predatorGrowth  = inputs.at(2)(4);
    predatorLoss    = inputs.at(2)(5);

  
  if (inputDimWrt == 1) {
    // create a vector to hold the ModPiece output, this will hold the time derivatives of population
    Eigen::MatrixXd output(2, 2);

    // compute the growth rates as we did in example-1
    output(0,0) =  -1.0*preyPop*preyGrowth/preyCapacity + (1.0-preyPop/preyCapacity)*preyGrowth - predPop*predationRate/(predationHunger+preyPop) + predPop*preyPop*predationRate/pow(predationHunger+preyPop,2.0);
    output(0,1) = -1.0 * (preyPop * predationRate) / (preyPop + predationHunger);

    output(1,0) = predPop*predationRate*predatorGrowth/(predationHunger+preyPop) - predPop*preyPop*predationRate*predatorGrowth/pow(predationHunger+preyPop,2.0);
    output(1,1) = (preyPop * predatorGrowth * predationRate) / (preyPop + predationHunger) - predatorLoss;

    // return the output
    return output;
  } else if (inputDimWrt == 2) {
    Eigen::MatrixXd output(2, 6);

    output(0, 0) = preyPop * (1-preyPop / preyCapacity);
    output(0, 1) = (pow(preyPop, 2.0) * preyGrowth) / pow(preyCapacity, 2.0);
    output(0, 2) = -(predPop * preyPop) / (preyPop + predationHunger);
    output(0, 3) = (predPop * preyPop * predationRate) / pow(preyPop + predationHunger, 2.0);
    output(0, 4) = 0.0;
    output(0, 5) = 0.0;

    output(1, 0) = 0.0;
    output(1, 1) = 0.0;
    output(1, 2) = predPop*preyPop*predatorGrowth/(preyPop+predationHunger);
    output(1, 3) = -(predPop * preyPop * predatorGrowth * predationRate) / pow(preyPop + predationHunger, 2.0);
    output(1, 4) = preyPop*predPop*predationRate/(predationHunger+preyPop);
    output(1, 5) = -predPop;

    return output;
  } else {
    assert(inputDimWrt < inputSizes.size());
    return Eigen::MatrixXd();
  }
}

std::shared_ptr<ModPiece> ecology::CreatePredPreyModel(double totalTime, int numObservations){

	// create the ode right hand side
    auto rhs = make_shared<PredPreyRHS>();

    // create a parameter list to hold the ODE solver parameters
    boost::property_tree::ptree odeParams;

    odeParams.put("CVODES.ATOL", 1e-5);           // absolute tolerance
    odeParams.get("CVODES.RTOL", 1e-5);           // relative tolerance
    odeParams.put("CVODES.Method", "BDF");         // either BDF or Adams
    odeParams.put("CVODES.SolveMethod", "Newton"); // either Newton or Iter
    odeParams.put("CVODES.MaxStepSize", 8e-1);     // maximum step size, should be small for accurate discrete adjoint
    odeParams.put("CVODES.PreferAdjoint",false);    // use adjoint or forward sensitivities
	
    // create a vector of times where we want to observe the ODE state
    Eigen::VectorXd obsTimes;
    obsTimes.setLinSpaced(numObservations, 0.0, totalTime);

    // create the ODE model
    return make_shared<OdeModPiece>(rhs, obsTimes, odeParams);
}

double ecology::EquilibriumPredicate(Eigen::VectorXd const& inputs){
	
    double a = inputs(0);
    double K = inputs(1);
    double c = inputs(2);
    double b = inputs(3);
    double g = inputs(4);
    double d = inputs(5);
	
	// compute the equilibrium point
	double P = d*b/(c*g-d);
	double Q = -a*(P*P+(b-K)*P-b*K)/(c*K);
    
	return std::min<double>(P,Q);
}



double ecology::StabilityPredicate(Eigen::VectorXd const& inputs){
	
    double a = inputs(0);
    double K = inputs(1);
    double c = inputs(2);
    double b = inputs(3);
    double g = inputs(4);
    double d = inputs(5);
	
	// compute the equilibrium point
	double P = d*b/(c*g-d);
	double Q = -a*(P*P+(b-K)*P-b*K)/(c*K);
	
	// build the jacobian at the equilibrium point
	Eigen::Matrix<double,2,2> jac(2,2);
	jac(0,0) = a * (1 - P / K) - a * P / K - c * Q / (b + P) + c * P * Q * pow(b + P, -2.0);
	jac(0,1) = -c * P / (b + P);
	jac(1,0) = g * c * Q / (b + P) - g * c * P * Q * pow(b + P, -2.0);
	jac(1,1) = g * c * P / (b + P) - d;
	
	
	// compute the eigenvalues of the jacobian
	auto lambdas = jac.eigenvalues();
	
	// return a result that is positive when both lambdas are positive, and zero otherwise
	double real1 = lambdas(0).real();
	double real2 = lambdas(1).real();
	
	//std::cout << inputs.transpose() << std::endl;
	//std::cout << real1 << " and " << real2 << std::endl;
	return std::min<double>(real1,real2);
}