#include <boost/property_tree/ptree.hpp>

#ifdef MUQ_MPI
# include <mpi.h>
# include <boost/mpi.hpp>
namespace mpi = boost::mpi;
#endif // if MUQ_MPI


#include "MUQ/Inference/TransportMaps/TransportMap.h"
#include "MUQ/Inference/TransportMaps/MapFactory.h"
#include "MUQ/Inference/TransportMaps/LinearBasis.h"
#include "MUQ/Inference/TransportMaps/StaticPolynomials.h"
#include "MUQ/Inference/TransportMaps/PolynomialBasis.h"
#include "MUQ/Inference/TransportMaps/BasisFactory.h"

#include "MUQ/Utilities/HDF5Wrapper.h"
#include "MUQ/Utilities/RandomGenerator.h"

#include "MUQ/Inference/MCMC/SingleChainMCMC.h"
#include "MUQ/Inference/MCMC/TransportMapKernel.h"
#include "MUQ/Inference/MAP/MAPbase.h"

#include "MUQ/Modelling/GaussianPair.h"
#include "MUQ/Modelling/EmpiricalRandVar.h"
#include "MUQ/Modelling/ModGraph.h"

#include "MUQ/Modelling/UniformDensity.h"
#include "MUQ/Modelling/VectorPassthroughModel.h"
#include "MUQ/Modelling/DensityProduct.h"
#include "MUQ/Modelling/ModPieceDensity.h"
#include "MUQ/Modelling/ModParameter.h"
#include "MUQ/Modelling/SliceModel.h"
#include "MUQ/Modelling/LinearModel.h"

#include "MUQ/Inference/ProblemClasses/SamplingProblem.h"
#include "MUQ/Inference/ProblemClasses/InferenceProblem.h"
#include "MUQ/Inference/ProblemClasses/GaussianFisherInformationMetric.h"

#include "MUQ/Utilities/JobManager.h"

#include <boost/timer/timer.hpp>

// Ecology related includes
#include "PopulationModels.h"
#include "../../MUQ_INSTALL/muq_external/include/boost/concept_check.hpp"

using namespace ecology;
using namespace muq::Modelling;
using namespace muq::Inference;
using namespace muq::Utilities;
using namespace std;


shared_ptr<ModGraph> BuildProblem(string groupName){
  
  const int numObservations = 5;
  const double endTime = 50;
  
  
  Eigen::VectorXd trueParams(8);
  trueParams << 50.0, // initial prey population
  5.0,  // initial predator population
  0.6,  // prey growth
  100.0,// prey carrying capacity
  1.2,  // predation rate
  25.0, // predator hunger
  0.5,  // predator growth
  0.3;  // predator death
  
  Eigen::MatrixXd scalingMat = trueParams.asDiagonal();
  
  
  // Set up the forward model portion of the graph
  shared_ptr<ModGraph> sys = make_shared<ModGraph>();
  sys->AddNode(make_shared<LinearModel>(scalingMat),"inferenceTarget");
  
  sys->AddNode(make_shared<SliceModel>(8,2,7),"jointParams");
  sys->AddEdge("inferenceTarget","jointParams",0);
  
  sys->AddNode(CreatePredPreyModel(endTime, numObservations),"forwardModel");
  
  sys->AddNode(make_shared<SliceModel>(8,0,1),"initialPops");
  sys->AddEdge("inferenceTarget","initialPops",0);
  
  sys->AddEdge("initialPops","forwardModel",0);
  sys->AddEdge("jointParams","forwardModel",1);
  
  // Set up the prior, likelihood, and posterior portion of the graph
  
  
  // generate data using the priorMode
  auto forwardMod = ModGraphPiece::Create(sys,"forwardModel");
  
  Eigen::VectorXd data,errorVar;
  
  data.resize(2*numObservations);
  data << 50.8527702127573917, 7.49368421385178074, 49.9344084804869937, 29.0076973055928029, 10.4141295786439763, 8.52740374865942741, 81.6668751648996363, 12.1709127652407982, 12.0051119275026075, 21.572288457234599;
  
  // add some noise to the data
  errorVar.resize(2*numObservations);
  for(int i=0 ; i<2*numObservations; i+=2){
    errorVar(i) = 10.0; // prey population observation error
    errorVar(i+1) = 10.0; // predator population observation error
  }
  
  HDF5Wrapper::WriteMatrix(groupName+"/ObservationData",data);
  Eigen::VectorXd obsTimes(numObservations);
  obsTimes.setLinSpaced(numObservations, 0.0, endTime);
  HDF5Wrapper::WriteMatrix(groupName+"/ObservationTimes",obsTimes);
  
  
  sys->AddNode(make_shared<UniformDensity>(make_shared<UniformUserPredicate>(ecology::EquilibriumPredicate,6)),"equilibriumPrior");
  sys->AddEdge("jointParams","equilibriumPrior",0);
  
  sys->AddNode(make_shared<UniformDensity>(make_shared<UniformUserPredicate>(ecology::StabilityPredicate,6)),"stabilityPrior");
  sys->AddEdge("jointParams","stabilityPrior",0);
  
  Eigen::VectorXd lb = 0.001*trueParams;
  Eigen::VectorXd ub = 50.0*trueParams;
  
  sys->AddNode(make_shared<UniformDensity>(make_shared<UniformBox>(lb,ub)),"boxPrior");
  sys->AddEdge("inferenceTarget","boxPrior",0);
  
  sys->AddNode(make_shared<DensityProduct>(3),"prior");
  sys->AddEdge("equilibriumPrior","prior",0);
  sys->AddEdge("stabilityPrior","prior",1);
  sys->AddEdge("boxPrior","prior",2);
  
  sys->AddNode(make_shared<GaussianDensity>(data,errorVar),"likelihood");
  sys->AddEdge("forwardModel","likelihood",0);
  
  // create the posterior 
  sys->AddNode(make_shared<DensityProduct>(2), "posterior");
  sys->AddEdge("prior","posterior",0);
  sys->AddEdge("likelihood","posterior",1);
  
  sys->writeGraphViz("PredPreyModelDerivs.pdf",ModGraph::DERIV_COLOR,true);
  
  return sys;
}


void ReadParams(const string &filename, boost::property_tree::ptree &params){
  
  // open a filestream
  ifstream fin(filename.c_str());
  boost::property_tree::read_xml(fin,params);
  fin.close();
}

void WriteResults(const string &hdf5Root, boost::property_tree::ptree &params, shared_ptr<EmpiricalRandVar> samples, shared_ptr<ModGraph> sys, double runTime){
  
  std::cout << "Writing samples into " << hdf5Root << std::endl;
  samples->WriteHDF5(hdf5Root+ "/chain");
  
  HDF5Wrapper::WriteMatrix<double,1,1>(hdf5Root + "/evals", Eigen::VectorXd::Constant(1,sys->GetNodeModel("forwardModel")->GetNumCalls("Evaluate")));
  HDF5Wrapper::WriteMatrix<double,1,1>(hdf5Root + "/gradEvals", Eigen::VectorXd::Constant(1,sys->GetNodeModel("forwardModel")->GetNumCalls("Gradient")));
  HDF5Wrapper::WriteMatrix<double,1,1>(hdf5Root + "/jacEvals", Eigen::VectorXd::Constant(1,sys->GetNodeModel("forwardModel")->GetNumCalls("Jacobian")));
  HDF5Wrapper::WriteMatrix<double,1,1>(hdf5Root + "/runtime", Eigen::VectorXd::Constant(1,runTime));
  
  
  // Now, make prior prediction samples (plotting much more often though)
  const int numObs = 100;
  const double endTime = 80.0;
  
  sys->RemoveNode("forwardModel");
  sys->AddNode(CreatePredPreyModel(endTime, numObs),"forwardModel");
  sys->AddEdge("initialPops","forwardModel",0);
  sys->AddEdge("jointParams","forwardModel",1);
  
  auto forwardMod = ModGraphPiece::Create(sys,"forwardModel");
  
  
  const int numPredicts = 100;
  Eigen::MatrixXd preds(2*numObs,numPredicts);
  
  std::cout << "\nSampling posterior predictive...\n";
  for(int i=0; i<numPredicts; ++i)
    preds.col(i) = forwardMod->Evaluate(samples->Sample());
  
  HDF5Wrapper::WriteMatrix(hdf5Root + "/PredictionData",preds);
  
  Eigen::VectorXd predTimes(numObs);
  predTimes.setLinSpaced(numObs,0.0,endTime);
  HDF5Wrapper::WriteMatrix(hdf5Root + "/PredictionTimes",predTimes);
}


string GetHdf5Root(boost::property_tree::ptree const& params)
{
  std::stringstream ss;
  bool hasCopyNumber = params.get("CopyNumber",-1)!=-1;
  
  if(hasCopyNumber)
    ss << std::setfill ('0') << std::setw (3) << params.get<int>("CopyNumber");
  
  string copyNum = ss.str();
  
  string rootName;
  string problemName = params.get("ProblemName","");
  string kernelName = params.get<std::string>("MCMC.Kernel");
  
  if(problemName.compare("")!=0)
    rootName = "/" + problemName;
  
  if(kernelName.compare("TransportMap")==0)
    rootName += "/PolyOrder" + params.get<std::string>("MCMC.TransportMap.PolyOrder");
  
  rootName += "/" + kernelName;
  
  if(kernelName.compare("MHKernel")==0)
    rootName += "/" + params.get<std::string>("MCMC.Proposal");
  
  if(kernelName.compare("DR")==0)
    rootName += "/" + params.get<std::string>("MCMC.DR.ProposalSpecification");
  
  if(kernelName.compare("TransportMap")==0)
  {
    string subRoot = GetHdf5Root(params.get_child("MCMC.TransportMap.SubMethod"));
    rootName += subRoot;
  }
  
  if(hasCopyNumber)
    rootName += "/Trial" + copyNum;
  
  return rootName;
  
}

void RunPredPrey(boost::property_tree::ptree const& constParams){
  
  
  boost::property_tree::ptree params;
  
  params.put_child("MCMC",constParams.get_child("MCMC"));
  params.put("ProblemName","PredatorPrey");
  params.put("Verbose",3);
  params.put("CopyNumber",constParams.get<int>("CopyNumber"));
  
  string hdf5Root = GetHdf5Root(params);
  auto sys = BuildProblem(hdf5Root);
  
  std::shared_ptr<InferenceProblem> problem;
  if((params.get<string>("MCMC.Proposal").compare("MMALA")==0)||(params.get<string>("MCMC.TransportMap.SubMethod.MCMC.Proposal").compare("MMALA")==0))
  {
    // use the Fisher information metric
    auto likelihood = dynamic_pointer_cast<GaussianDensity>(sys->GetNodeModel("likelihood"));
    auto forwardModel = ModGraph::ConstructModPiece(sys,"forwardModel");
    auto metric = make_shared<GaussianFisherInformationMetric>(likelihood, forwardModel);
    problem = make_shared<InferenceProblem>(sys,metric);
  }
  else
  {
    problem = make_shared<InferenceProblem>(sys);
  }
  
  // the forward model scales the inputs, so the "true" parameter is a vector of ones
  Eigen::VectorXd x0 = Eigen::VectorXd::Ones(8);
  
  
  std::shared_ptr<MCMCBase> solver;
  if(params.get<std::string>("MCMC.Kernel").compare("TransportMap")){
    std::cout << "Using Standard MCMC approach\n";
    solver = MCMCBase::ConstructMCMC(problem, params);
    
  }else{
    
    // create a linear basis in every direction
    auto bigBasis = BasisFactory::TotalOrderHermite(8, params.get<int>("MCMC.TransportMap.PolyOrder"));
    auto map = MapFactory::CreateIdentity(bigBasis);
    
    std::cout << "Using TransportMap accelerated MCMC\n";
    auto TestKernel = make_shared<TransportMapKernel>(problem,params,map);
    solver = make_shared<SingleChainMCMC>(TestKernel,params);
  }
  
  auto mapSolver = MAPbase::Create(problem,params);
  
  boost::timer::cpu_timer timer;
  Eigen::VectorXd xc = mapSolver->Solve(x0);
  boost::timer::cpu_times elapsed = timer.elapsed();
  double mapTime = double(elapsed.user + elapsed.system)/1e6; // cpu time in milliseconds
  std::cout << "MAP Point = " << xc.transpose() << std::endl;
  std::cout << "Found map in " << mapTime << "(ms)" << std::endl;
  
  timer.start();
  auto mcmcSamps = solver->Sample(xc);
  
  elapsed = timer.elapsed();
  double mcmcTime = double(elapsed.user + elapsed.system)/1e6; // cpu time in milliseconds
  
  std::cout << "Number of evaluations = " << sys->GetNodeModel("forwardModel")->GetNumCalls("Evaluate") << std::endl;
  
  WriteResults(hdf5Root, params, mcmcSamps, sys, mcmcTime); 	
  solver->WriteAttributes(hdf5Root); // needs to be called after groups have been created
  
  HDF5Wrapper::FlushFile();	
}

REGISTER_JOB_FUNCTION(RunPredPrey, GetHdf5Root)

int main(int argc, char **argv){
  
  #ifdef MUQ_MPI
  std::cout << "Intializing MUQ_MPI...\n";
  boost::mpi::environment env(argc, argv);
  #endif // ifdef MUQ_MPI
  
  std::vector<string> MCMCKernels, TransportMapKernels, TransportMapProposals, MCMCProposals;
  
  // job 1 -- DRAM
  MCMCKernels.push_back("DR");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 2 -- AM
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("AM");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 3 -- TMDR (Global local)
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DR");
  TransportMapProposals.push_back("GlobalProposal,LocalProposal2");
  
  // job 4 -- TMDR (Local Local)
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DR");
  TransportMapProposals.push_back("LocalProposal2,LocalProposal3");
  
  // job 5 -- TMMH
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("MHProposal");  
  
  // job 6 -- TMMIX
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("MixtureProposal");
  
  // job 7 -- AMALA
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("AMALA");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 8 -- smMALA
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("MMALA");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
   
  //job 9 -- TMLA
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("PreMALA");
  
  
  const int numRepeats = 30;
  
  boost::property_tree::ptree jobConfig;
  
  /////////////////////////////////////////////////////
  // set up all the BOD jobs
  boost::property_tree::ptree localJob;
  localJob.put("NumberOfCopies", numRepeats);
  localJob.put("Function", "RunPredPrey");
  
  
  // open up the hdf5 file
  HDF5Wrapper::OpenFile("results/PredPreyResults.h5");
  
  boost::property_tree::ptree mcmcParams;
  ReadParams("inputParams/PredPreyParams.xml",mcmcParams);
  
  for(int i=0; i<MCMCKernels.size(); ++i){
    
    mcmcParams.put("MCMC.Kernel",MCMCKernels.at(i));
    mcmcParams.put("MCMC.Proposal",MCMCProposals.at(i));
    mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.Kernel",TransportMapKernels.at(i));
    
    if(TransportMapKernels.at(i).compare("DR")==0){
      mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.DR.ProposalSpecification",TransportMapProposals.at(i));
    }else{
      mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.Proposal",TransportMapProposals.at(i));	
    }
    
    if(MCMCKernels.at(i).compare("TransportMap")==0){  
      mcmcParams.put("MCMC.TransportMap.PolyOrder",3);
      //mcmcParams.put("MCMC.TransportMap.AdaptScale",10000.0);
      localJob.put_child("Data",mcmcParams);
      jobConfig.add_child("Job",localJob);
    }else{
      localJob.put_child("Data",mcmcParams);
      jobConfig.add_child("Job",localJob);
    }
  }
  JobManager::RunJobs(jobConfig); 
  
  std::cout << "About to enter CloseFile()" << std::endl;
  // close the results file
  HDF5Wrapper::CloseFile();      
  
  
  return 0;
}
