

#include <boost/property_tree/ptree.hpp>

#ifdef MUQ_MPI
# include <mpi.h>
# include <boost/mpi.hpp>
namespace mpi = boost::mpi;
#endif // if MUQ_MPI


#include "MUQ/Inference/TransportMaps/TransportMap.h"
#include "MUQ/Inference/TransportMaps/MapFactory.h"
#include "MUQ/Inference/TransportMaps/LinearBasis.h"
#include "MUQ/Inference/TransportMaps/StaticPolynomials.h"
#include "MUQ/Inference/TransportMaps/PolynomialBasis.h"
#include "MUQ/Inference/TransportMaps/BasisFactory.h"

#include "MUQ/Utilities/HDF5Wrapper.h"
#include "MUQ/Utilities/RandomGenerator.h"

#include "MUQ/Inference/MCMC/SingleChainMCMC.h"
#include "MUQ/Inference/MCMC/TransportMapKernel.h"

#include "MUQ/Inference/MAP/MAPbase.h"

#include "MUQ/Modelling/GaussianPair.h"
#include "MUQ/Modelling/EmpiricalRandVar.h"
#include "MUQ/Modelling/ModGraph.h"

#include "MUQ/Modelling/UniformDensity.h"
#include "MUQ/Modelling/VectorPassthroughModel.h"
#include "MUQ/Modelling/DensityProduct.h"
#include "MUQ/Modelling/ModPieceDensity.h"


#include "MUQ/Inference/ProblemClasses/SamplingProblem.h"
#include "MUQ/Inference/ProblemClasses/InferenceProblem.h"
#include "MUQ/Inference/ProblemClasses/GaussianFisherInformationMetric.h"


#include "MUQ/Utilities/JobManager.h"

#include <boost/timer/timer.hpp>

// include my set of simple test densities
#include "SimpleTestDensities.h"

/// namespaces
using namespace muq::Modelling;
using namespace muq::Inference;
using namespace muq::Utilities;
using namespace std;

void ReadParams(const string &filename, boost::property_tree::ptree &params){
  // open a filestream
  ifstream fin(filename.c_str());
  boost::property_tree::read_xml(fin,params);
  fin.close();
}

void WriteTrueDensity(const string &rootName, shared_ptr<Density> dens, const Eigen::MatrixXd &bounds){
  
  // I only need to plot 2d problems
  assert(bounds.rows()==2);
  assert(bounds.cols()==2);
  
  const int Nx = 80;
  const int Ny = 80;
  
  const double dx = (bounds(0,1)-bounds(0,0))/(Nx-1);
  const double dy = (bounds(1,1)-bounds(1,0))/(Ny-1);
  
  Eigen::MatrixXd densEvals(Nx,Ny);
  
  for(int j=0; j<Ny; ++j){
    for(int i=0; i<Nx; ++i){
      Eigen::VectorXd loc(2);
      loc << bounds(0,0) + dx*i, bounds(1,0)+dy*j;
      densEvals(i,j) = dens->LogDensity(loc);
    }
  }
  std::string dataset = rootName + "/trueDensity";
  HDF5Wrapper::WriteMatrix<double>(dataset,densEvals);
}

void WriteMapDensity(const string &rootName, boost::property_tree::ptree &params, shared_ptr<TransportMap> map, const Eigen::MatrixXd &bounds){
  
  // I only need to plot 2d problems
  assert(bounds.rows()==2);
  assert(bounds.cols()==2);
  
  const int Nx = 80;
  const int Ny = 80;
  
  const double dx = (bounds(0,1)-bounds(0,0))/(Nx-1);
  const double dy = (bounds(1,1)-bounds(1,0))/(Ny-1);
  
  Eigen::MatrixXd densEvals(Nx,Ny);
  
  for(int j=0; j<Ny; ++j){
    for(int i=0; i<Nx; ++i){
      Eigen::VectorXd loc(2);
      loc << bounds(0,0) + dx*i, bounds(1,0)+dy*j;
      densEvals(i,j) = -0.5*map->Evaluate(loc).squaredNorm() + map->Jacobian(loc).diagonal().array().log().sum();
    }
  }
  
  std::string dataset = rootName + "/mapDensity";
  HDF5Wrapper::WriteMatrix(dataset,densEvals);
}


void WriteRefDensity(const string &rootName, shared_ptr<Density> dens, boost::property_tree::ptree &params, shared_ptr<TransportMap> map, const Eigen::MatrixXd &bounds, const Eigen::VectorXd &x0){
  
  // I only need to plot 2d problems
  assert(bounds.rows()==2);
  assert(bounds.cols()==2);
  
  const int Nx = 80;
  const int Ny = 80;
  
  const double dx = (bounds(0,1)-bounds(0,0))/(Nx-1);
  const double dy = (bounds(1,1)-bounds(1,0))/(Ny-1);
  
  Eigen::MatrixXd densEvals(Nx,Ny);
  
  for(int j=0; j<Ny; ++j)
  {
    for(int i=0; i<Nx; ++i)
    {
      Eigen::VectorXd refLoc(2);
      refLoc << bounds(0,0) + dx*i, bounds(1,0)+dy*j;
      Eigen::VectorXd targetLoc = map->EvaluateInverse(refLoc,x0);
      densEvals(i,j) = dens->LogDensity(targetLoc) - map->Jacobian(targetLoc).diagonal().array().log().sum();
    }
  }
  
  std::string dataset = rootName + "/refMapDensity";
  HDF5Wrapper::WriteMatrix(dataset,densEvals);
  
}

void WriteResults(const string &datasetRoot, boost::property_tree::ptree &params, shared_ptr<EmpiricalRandVar> samples, int numEvals, int numGradEvals, double runTime){
  
  std::cout << "Writing results into " << datasetRoot << std::endl;
  samples->WriteHDF5(datasetRoot+ "/chain");
  
  HDF5Wrapper::WriteMatrix<double,Eigen::Dynamic,1>(datasetRoot + "/evals", Eigen::VectorXd::Constant(1,numEvals));
  HDF5Wrapper::WriteMatrix<double,Eigen::Dynamic,1>(datasetRoot + "/gradEvals", Eigen::VectorXd::Constant(1,numGradEvals));
  HDF5Wrapper::WriteMatrix<double,Eigen::Dynamic,1>(datasetRoot + "/runtime", Eigen::VectorXd::Constant(1,runTime));
}

string GetHdf5Root(boost::property_tree::ptree const& params)
{
  std::stringstream ss;
  bool hasCopyNumber = params.get("CopyNumber",-1)!=-1;
  
  if(hasCopyNumber)
    ss << std::setfill ('0') << std::setw (3) << params.get<int>("CopyNumber");
  
  string copyNum = ss.str();
  
  string rootName;
  string problemName = params.get("ProblemName","");
  string kernelName = params.get<std::string>("MCMC.Kernel");
  
  if(problemName.compare("")!=0)
    rootName = "/" + problemName;
  
  if((problemName.compare("BOD")==0)&&(kernelName.compare("TransportMap")==0))	
    rootName += "/PolyOrder" + params.get<std::string>("MCMC.TransportMap.PolyOrder");
  
  rootName += "/" + kernelName;
  
  if(kernelName.compare("MHKernel")==0)
    rootName += "/" + params.get<std::string>("MCMC.Proposal");
  
  if(kernelName.compare("DR")==0)
    rootName += "/" + params.get<std::string>("MCMC.DR.ProposalSpecification");
  
  if(kernelName.compare("TransportMap")==0)
  {
    string subRoot = GetHdf5Root(params.get_child("MCMC.TransportMap.SubMethod"));
    rootName += subRoot;
  }
  
  if(hasCopyNumber)
    rootName += "/Trial" + copyNum;
  
  return rootName;
  
}


void RunBODModel(boost::property_tree::ptree const& constParams){
  
  assert(constParams.get("CopyNumber",-1)!=-1);
  
  boost::property_tree::ptree params;
  params.put_child("MCMC",constParams.get_child("MCMC"));
  params.put_child("Opt",constParams.get_child("Opt"));
  params.put("ProblemName","BOD");
  params.put("Verbose",3);
  params.put("CopyNumber",constParams.get<int>("CopyNumber"));
  
  if((params.get<string>("MCMC.TransportMap.SubMethod.MCMC.Kernel").compare("MHKernel")==0)&&(params.get<string>("MCMC.TransportMap.SubMethod.MCMC.Proposal").compare("PreMALA")==0))
    params.put("MCMC.TransportMap.AdaptScale",2);
  
  // get the hdf root group
  string hdf5Root = GetHdf5Root(params);
  
  std::cout << "----------------------------\nRunning BOD-model problem...\n";
  
  // generate the "data"
  const int numData = 20;
  const double obsVar = 2e-4;
  
  Eigen::VectorXd dataLocs(numData);
  dataLocs.setLinSpaced(numData,1,5);
  
  // get the "exact data"
  Eigen::VectorXd dataVals(numData);
  
  dataVals << 0.098976286293865470278, 0.12516476737942788811, 0.13261159285271648112, 0.1511007498356576606, 0.18310406794808162245, 0.19551644332053619291, 0.21939982217855563684, 0.21435760337843853818, 0.26830252993288139507, 0.25246633404963125802, 0.26602208797213733327, 0.29800368899781060072, 0.31123858155378536505, 0.29617078526513229697, 0.33605678216467343988, 0.320286329267592762, 0.34758826941454601389, 0.35821494031148093384, 0.3706300019795769618, 0.39623541337378903604;

  
  // set up the problem
  std::shared_ptr<ModGraph> graph = make_shared<ModGraph>();
  graph->AddNode(std::make_shared<VectorPassthroughModel>(2), "inferenceTarget");
  
  auto bodMod = make_shared<BODModel>(dataLocs);
  graph->AddNode(bodMod,"forwardModel");
  
  auto errorMod = std::make_shared<GaussianDensity>(dataVals, obsVar);
  graph->AddNode(errorMod, "likelihood");
  graph->AddNode(std::make_shared<UniformDensity>(std::make_shared<UniformBox>(Eigen::VectorXd::Constant(2, -1000),
                                                                               Eigen::VectorXd::Constant(2, 1000))), "prior");
  graph->AddNode(std::make_shared<DensityProduct>(2), "posterior");
  
  graph->AddEdge("inferenceTarget", "forwardModel", 0);
  graph->AddEdge("forwardModel", "likelihood", 0);
  graph->AddEdge("inferenceTarget", "prior", 0);
  graph->AddEdge("likelihood", "posterior", 0);
  graph->AddEdge("prior", "posterior", 1);
  
  std::cout << "Done constructing graph\n";
  
  auto bodDens = ModGraph::ConstructDensity(graph, "posterior");
  
  std::cout << "Created bodDens\n";
  
  std::shared_ptr<InferenceProblem> bodProblem;
  if((params.get<string>("MCMC.Proposal").compare("MMALA")==0)||(params.get<string>("MCMC.TransportMap.SubMethod.MCMC.Proposal").compare("MMALA")==0))
  {
    // use the Fisher information metric
    auto metric = make_shared<GaussianFisherInformationMetric>(errorMod, bodMod);
    bodProblem = make_shared<InferenceProblem>(graph,metric);
  }
  else
  {
    bodProblem = make_shared<InferenceProblem>(graph);
  }
  
  std::cout << "Done creating problem\n";
  
  // build the initial map, make it the identity
  //auto bases = BasisFactory::TotalOrderHermite(2, 1);
  auto bases = BasisFactory::TotalOrderHermite(2, params.get("MCMC.TransportMap.PolyOrder",3));
  auto initialMap = MapFactory::CreateIdentity(bases);
  
  
  std::shared_ptr<MCMCBase> solver;
  if(params.get<std::string>("MCMC.Kernel").compare("TransportMap"))
  {
    std::cout << "Using Normal MCMC method\n";
    solver = MCMCBase::ConstructMCMC(bodProblem, params);
  }
  else
  {  
    std::cout << "Using TransportMap accelerated MCMC\n";
    auto TestKernel = make_shared<TransportMapKernel>(bodProblem,params,initialMap);
    solver = make_shared<SingleChainMCMC>(TestKernel,params);
  }
  
  // sample the distribution
  Eigen::VectorXd x0(2);
  x0 << 0.9294,   0.1040;
  boost::timer::cpu_timer timer;
  
  // find the map point 
  auto mapSolver = MAPbase::Create(bodProblem,params);
  Eigen::VectorXd xc = mapSolver->Solve(x0);
  std::cout << "Found map at: " << xc.transpose() << std::endl;
  
  auto mcmcSamps = solver->Sample(xc);
  
  boost::timer::cpu_times elapsed = timer.elapsed();
  double runTime = double(elapsed.user + elapsed.system)/1e6; // cpu time in milliseconds
  
  
  WriteResults(hdf5Root, params, mcmcSamps, bodMod->GetNumEvals(),bodMod->GetNumGradEvals(),runTime);
  solver->WriteAttributes(hdf5Root); // needs to be called after groups have been created
  
  if(!params.get<std::string>("MCMC.Kernel").compare("TransportMap"))
  {
    // write the true and map induced densities
    Eigen::MatrixXd bounds(2,2);
    bounds << 0.73,1.6,
    0.05, 0.15;
    //WriteTrueDensity("BOD",bodDens,bounds);
    WriteMapDensity(hdf5Root,params,initialMap,bounds);
    
    bounds << -3.5, 3.5, -3.5, 3.5;
    WriteRefDensity(hdf5Root,bodDens,params,initialMap, bounds,mcmcSamps->Sample());
  }
  HDF5Wrapper::FlushFile();
  //HDF5Logging::WipeBuffers();
};

REGISTER_JOB_FUNCTION(RunBODModel, GetHdf5Root)

int main(int argc, char **argv){
  
  #ifdef MUQ_MPI
  std::cout << "Intializing MUQ_MPI...\n";
  boost::mpi::environment env(argc, argv);
  #endif // ifdef MUQ_MPI
  
  std::vector<string> MCMCKernels, MCMCProposals,TransportMapKernels, TransportMapProposals;
  
  // job 1 -- NUTS
  MCMCKernels.push_back("StanNUTS");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 2 -- DRAM
  MCMCKernels.push_back("DR");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 3 -- AM
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("AM");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 4 -- AMALA
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("AMALA");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 5 -- smMALA
  MCMCKernels.push_back("MHKernel");
  MCMCProposals.push_back("MMALA");
  TransportMapKernels.push_back("DUMMY");
  TransportMapProposals.push_back("DUMMY");
  
  // job 6 -- TMDR with global proposal
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DR");
  TransportMapProposals.push_back("GlobalProposal,LocalProposal1");
  
  //job 7 -- TMDR with local proposal
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("DR");
  TransportMapProposals.push_back("LocalProposal2,LocalProposal3");
  
  // job 8 -- TMMIX
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("MixtureProposal");
  
  //job 9 -- TMLA
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("PreMALA");
  
  // job 10 -- TMRW
  MCMCKernels.push_back("TransportMap");
  MCMCProposals.push_back("DUMMY");
  TransportMapKernels.push_back("MHKernel");
  TransportMapProposals.push_back("MHProposal");
  
  const int numRepeats = 30;
  
  boost::property_tree::ptree jobConfig;
  
  /////////////////////////////////////////////////////
  // set up all the BOD jobs
  boost::property_tree::ptree localJob;
  localJob.put("NumberOfCopies", numRepeats);
  localJob.put("Function", "RunBODModel");
  
  Eigen::VectorXi porders(1);
  porders << 3;
  
  Eigen::VectorXd adaptScales(1);
  adaptScales << 10000;
  
  // open up the hdf5 file
  HDF5Wrapper::OpenFile("results/BODResults.h5");
  
  for(int i=0; i<MCMCKernels.size(); ++i){
    
    boost::property_tree::ptree mcmcParams;
    ReadParams("inputParams/BodParams.xml",mcmcParams);
    
    mcmcParams.put("MCMC.Kernel",MCMCKernels.at(i));
    mcmcParams.put("MCMC.Proposal",MCMCProposals.at(i));
    
    // if this is a transport map, add a job for each polynomial order
    if(TransportMapKernels.at(i).compare("DUMMY")){
      
      mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.Kernel",TransportMapKernels.at(i));
      
      if(TransportMapKernels.at(i).compare("DR")==0){
        mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.DR.ProposalSpecification",TransportMapProposals.at(i));
      }else{
        mcmcParams.put("MCMC.TransportMap.SubMethod.MCMC.Proposal",TransportMapProposals.at(i));	
      }
      
      if((TransportMapKernels.at(i).compare("MHKernel")==0)&&(TransportMapProposals.at(i).compare("PreMALA")==0)){
        std::cout  << "Changing from "  << mcmcParams.get<double>("MCMC.TransportMap.AdaptScale");
        mcmcParams.put("MCMC.TransportMap.AdaptScale",1);
        std::cout << "  to " << mcmcParams.get<double>("MCMC.TransportMap.AdaptScale") << std::endl;
      }
      for(int p=0;p<porders.size();++p){
        mcmcParams.put("MCMC.TransportMap.PolyOrder",porders(p));
        mcmcParams.put("MCMC.TransportMap.AdaptScale",adaptScales(p));
        localJob.put_child("Data",mcmcParams);
        jobConfig.add_child("Job",localJob);	
      }
      
      
    }else{
      localJob.put_child("Data",mcmcParams);
      jobConfig.add_child("Job",localJob);
    }
  }
  
  JobManager::RunJobs(jobConfig); 
  
  std::cout << "About to enter CloseFile()" << std::endl;
  // close the results file
  HDF5Wrapper::CloseFile();      
  
  
  return 0;
}
